
// emacs: this is -*- c++ -*-
//
//   @file    nnlo_interface.cxx
//
//            grid filling routines for interface with
//            the nnlojet code from CITATION
//
//   Copyright (C) 2015 M.Sutton (sutt@cern.ch)
//
//   $Id: nnlo_interface.h, v0.0  Thu 4 Jul 2015 10:25:18 CEST sutt $


#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <cmath>

#include <sys/stat.h>


#include "amconfig.h"

#ifdef HAVE_APPL
// applgrid headers
#include "appl_grid/appl_grid.h"
#include "appl_grid/lumi_pdf.h"
#endif

#ifdef HAVE_FASTNLO
/// fastnlo headers ...
#include "fastnlotk/fastNLOCreate.h"
#include "fastnlotk/read_steer.h"
#endif

// nnlo interface header files
#include "nnlo_interface.h"
#include "nnlo_utils.h"
#include "nnlo_common.h"


// external hook functions
extern void (*appl_initptr)( const int& id, const std::string& gridname,
                             const int& nbins, const double& lo, const double& hi  );

extern void (*appl_fillptr)( const int& id, const double& obs, const double* wt );

extern void (*appl_fillrefptr)( const int& id,
                                const double& obs,
                                const double* wt );

extern void (*appl_termptr)( const int& id );

extern "C" void evolvepdf_(const double& x, const double& Q, double* xf);
extern "C" double alphaspdf_(const double& Q);


/// Declare the input and output grids
/// Declare grid-specific functions
static std::string _gridname; //  = "nnlojet.root";

#ifdef HAVE_APPL
//static std::int  _gridname; //  = "nnlojet.root";
static std::vector<appl::grid*> _grids;
static std::map<int, appl::grid*> _gridmap;
static std::map<int, std::string> _gridnames;

void nnlo::fill_appl_internal( appl::grid* _grid,
                               const int& order,
                               const double& x1, const double& x2, const double& Q2,
                               const double& obs, const double* wgt );

appl::grid* find( const int& id ) {
   std::map<int, appl::grid*>::iterator gitr = _gridmap.find(id);
   if ( gitr!=_gridmap.end() ) return gitr->second;
   std::cerr << "find() requested grid id " << id << " not found" << std::endl;
   return 0;
}
#endif

#ifdef HAVE_FASTNLO
static fastNLOCreate *ftableBase;
static fastNLOCreate *ftableNLO;
#endif


template<typename T>
std::ostream& operator<<( std::ostream& s, std::vector<T>& t ) {
  for ( size_t i=0 ; i<t.size() ; i++ ) s << "\t" << t[i];
  return s;
}


// external hook functions
extern "C" bool setup_nnlo_() {
  std::cout << "NNLOJET INFO: setting up hook functions ..." << std::endl;

  appl_initptr       = nnlo::init;
  appl_fillptr       = nnlo::fill;
  appl_fillrefptr    = nnlo::fill_ref;
  appl_termptr       = nnlo::term;

  return true;
}



/// Initialise the hook functions
static bool _ready = setup_nnlo_();


std::string findname( const int& id ) {
  std::map<int, std::string>::iterator gitr = _gridnames.find(id);
  if ( gitr!=_gridnames.end() ) return gitr->second;
  return "";
}


/// banner
std::string nnlo::banner() {
    std::string version = PACKAGE_VERSION;
    std::string webpage = PACKAGE_URL;
    std::string banner  = "\n";
    banner += "**************************************************************************************\n";
    banner += "*   NNLOJET:        E.W.N.Glover, T.Morgan [arXiv:.....]                             *\n";
    banner += "*   nnlo-bridge: E.W.N.Glover, C. Gwenlan, T.A.Morgan, K.Rabbertz, M.R.Sutton   ...  *\n";
    banner += "*   APPLgrid:       T. Carli et al.    [arXiv:0911.2985]                             *\n";
    banner += "*   fastNLO:        D. Britzger K.Rabbertz, M. Wobisch. et al. [arXiv:...]           *\n";
    banner += "**************************************************************************************\n";
    banner += "\n";

    banner += "\n";
    banner += "Web page: " + webpage + "\n";
    banner += "\n";
    return banner;
}



bool first_run = true;



//
// Initialisation; book interpolation grid(s)
//
void nnlo::init( const int& id, const std::string& gridname,
                 const int& nbins, const double& lo, const double& hi ) {

   std::cout << banner() << std::endl;
   _gridname = gridname;
   std::cout << "### nnlo::init(): gridname is " << _gridname << std::endl;

   // Where does this come from? From NNLOJETS common?
   int leading_order = inppar_.njets;
   int nloops        = inppar_.nloop;

   // Define some binning
   std::vector<double> binning;
   binning.push_back(10.);
   binning.push_back(100.);
   binning.push_back(1000.);

   // Not sure about idea for "id"; most flexible solution would be to
   // allow an APPLGRID, fastNLO table, or both for each requested grid.
   // For the moment assume only one grid "id".
   // If id>0 --> APPLGRID, id<0 --> fastNLO, id==0 --> both.

   if ( id>=0 ) { // APPLGRID requested

#ifdef HAVE_APPL

      // Book APPLGRID

      std::cout << "### nnlo::init(): Book APPLGRID ..." << std::endl;

      if ( file_exists( _gridname ) ) {

         std::cout << "################### File exists ..." << std::endl;

         // Open the existing grid

         std::cout << "nnlo::init() ready " << _ready << std::endl;
         std::cout << "nnlo::init() reading grid " << _gridname << " ..." << std::endl;

         appl::grid* _grid = new appl::grid(_gridname);

         _gridnames.insert( std::map<int, std::string>::value_type( id, _gridname ) );
         _gridmap.insert(   std::map<int, appl::grid*>::value_type( id, _grid ) );

         _grid->optimise(true);

         first_run = false;

	 std::cout << *_grid << std::endl;

      } else {

         std::cout << "nnlo::init() booking appl-grid from scratch with name " << _gridname << " ..." << std::endl;

         /// Create a new grid - these should be set from some values
         /// from the NNLOJET code

         /// eta bins from eta0_a_gg5g0q_i1_n1_z2_q1_r001

         /// need to allow setting these from the fortran ???

         int    Nobs   = nbins;
         double obsmin = lo;
         double obsmax = hi;  //don't change this anymore !!!!!

         std::cout << "booking grid: " << Nobs << "\trange " << lo << " - " << hi << std::endl;

         int    NQ2     = 20;
         double Q2min   = 10;
         double Q2max   = 4*14000*14000;
         double Q2order = 4;

         int    Nx     = 40;
         double xmin   = 0.00001;
         double xmax   = 0.9999999;
         double xorder = 5;

         // int leading_order = inppar_.njets;
         // int nloops        = inppar_.nloop;

         std::cout << " njets: " <<  inppar_.njets << "\tnloops " << inppar_.nloop << std::endl;

         std::string genpdf = "ZJ-LO.config";

         std::cout << "creating grid " << genpdf << std::endl;

         //    lumi_pdf lumi( genpdf );
         //    lumi.write_latex( std::cout );

         //    double apramval=5.;
         //    appl::grid::transformvar(apramval);

         int nigrids = 0;

         if ( nloops == 0 ) nigrids = 0;
         if ( nloops == 1 ) nigrids = 1;
         if ( nloops == 2 ) nigrids = 2;

         appl::grid* _grid = new appl::grid( NQ2,  Q2min,  Q2max, Q2order,
                                             Nx,   xmin,   xmax,  xorder,
                                             Nobs, obsmin, obsmax,
                                             genpdf, leading_order, nigrids );

         // Use the reweighting function
         _grid->reweight(true);

         // Add documentation
         _grid->addDocumentation( banner() );

         _gridnames.insert( std::map<int, std::string>::value_type( id, _gridname ) );
         _gridmap.insert(   std::map<int, appl::grid*>::value_type( id, _grid ) );

      }

#else

      std::cerr << "### nnlo::init(): APPLGRID requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   if ( id<0 ) { // fastNLO requested

#ifdef HAVE_FASTNLO

      // Book fastNLO table

      //      using namespace std;
      std::cout << "### nnlo::init(): Book fastNLO table ..." << std::endl;

      lumi_pdf pdffnlo("ZJ-LO.config");

      // DB. enable compilation pdffnlo.removeDuplicates();
      pdffnlo.serialise();
      std::cout << pdffnlo << std::endl;

      std::cout << "fastNLO initialisation" << std::endl;
      const std::string steeringNameSpace = "fastNLObase.str";
      std::cout << "Steering name space set to: " << steeringNameSpace << std::endl;

      // Read parton lumi mappings
      //     READ_NS("ZJ-LO.config", steeringNameSpace);

      // Values passed here will overwrite a possible value in the steering

      // fastNLOCreate *ftableBase;
      // ftableBase = new fastNLOCreate("NNLOJETS_CMS_2011_S9086218.str",
      //                                steeringNameSpace,
      //                                false);
      // ftableBase->SetOrderOfAlphasOfCalculation(2);

      // appl::grid* _grid = new appl::grid( NQ2,  Q2min,  Q2max, Q2order,
      //                                     Nx,   xmin,   xmax,  xorder,
      //                                     Nobs, obsmin, obsmax,
      //                                     genpdf, leading_order, nigrids );

      std::cout << "FASTNLO: Book fastNLO interpolation grid." << std::endl;
      std::cout << "FASTNLO: Set parameters that are passed from NNLOJETS to fastNLO!" << std::endl;
      ADD_NS("LeadingOrder", leading_order, steeringNameSpace);
      ADDARRAY_NS("SingleDifferentialBinning", binning, steeringNameSpace);
      ftableBase = new fastNLOCreate("fastNLObase.str",
                                     steeringNameSpace,
                                     true);
      // ftableBase = new fastNLOCreate("fastNLObase.str",
      //                                "fastNLOwarm.txt",
      //                                true);
      std::cout << "FASTNLO: fastNLO grid booked." << std::endl;
      // if (ftableBase->GetIsWarmup()) {
      //    ftableNLO = NULL;
      // } else {
      //    // This is no warmup run, prepare to fill NLO events
      //    // As it is initialized from the same steering,
      //    // the NLO grid will use the same warmup values than the LO grid
      //    ftableNLO = new fastNLOCreate("fastNLObase.str",
      //                                  steeringNameSpace,
      //                                  false);
      //    ftableNLO->SetOrderOfAlphasOfCalculation(2);
      // }

#else

      std::cerr << "### nnlo::init(): fastNLO requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   std::cout << "### nnlo::init(): End Init ..." << std::endl;

}







//
// Fill the interpolation grid with the weights
//
void nnlo::fill( const int& id, const double& obs, const double* wt ) {

   //   std::cout << "### nnlo::fill(): Fill grid ..." << std::endl;

   /// nominal kinematics
   /// use common/parfrac/ rather than common/pdf/
   //  double x1 =  pdf_.xi1;
   //  double x2 =  pdf_.xi2;
   double x1 =  parfrac_.x1;
   double x2 =  parfrac_.x2;
   double Q2  =  scales_.rscale[0];
   //  double Q2 =  Q*Q;

   if ( id>=0 ) { // APPLGRID requested

#ifdef HAVE_APPL

      // Fill APPLGRID

      //      std::cout << "### nnlo::fill(): Fill APPLGRID ..." << std::endl;

      appl::grid* _grid = find( id );

      if ( _grid ) {

	std::cout << "id:  " << id << "\nx1:  " << x1 << "\nx2:  " << x2 << "\nQ2: " << Q2 << std::endl;
	
	std::cout << "obs:  " << obs << std::endl;

	//        std::cout << "wt: \n";
	//    for ( int i=0 ; i<7 ; i++ ) std::cout << "\t" << wt[i] << std::endl;
	
	//    std::cout << "nnlo::fill() found grid " << _grid << std::endl;
	
	/// find iproc if need be
	
	int iorder = 0;

	const lumi_pdf* pdf = dynamic_cast<const lumi_pdf*>( ( (const appl::grid*)_grid)->genpdf(iorder) );

	int iproc = currentprocess_.iproc;

	/// get input parton species, calculate subprocess index
	
	/// Fixme: just for the time being - sort out which order we
	///        are dealing with properly later
	
	
	/// use common/partid/ rather than common/idparton/ ??
	//    int  ip1 = partid_.ip1;
	//    int  ip2 = partid_.ip2;
	//    std::cout << "nnlo::fill() pdf " << pdf << "\tpartons " << ip1 << " " << ip2 << std::endl;

         int iprocess = pdf->decideSubProcess( iproc );

         std::cout << "nnlo::fill() iproc " << iproc << " " << iprocess << std::endl;

         if ( iprocess!=-1 ) {

	   std::vector<double> wgt(pdf->size(), 0);
      
	   wgt[iprocess] = wt[0];

	   //      std::vector<double> weights(7,0);
	   
	   //      for ( int j=7 ; j-- ; ) weights[1] = wt[j];
	   
	   /// NB: now pass the weights in as a parameter ...
	   /// chose which os the maxscl=3 weights, you actuall
	   /// want to use
            //      weights[iprocess] = gridweights_.gridwgt[0];
	   
            //      std::cout << "nnlo::fill() x1 " << x1 << "\tx2 " << x2 << "\tQ2 " << Q2 << std::endl;
	   
            //      fill_appl_internal( _grid, iorder, x1, x2, Q2, obs, &weights[0] );
	   fill_appl_internal( _grid, iorder, x1, x2, Q2, obs, &wgt[0] );
	   
         }

      }

#else

      std::cerr << "### nnlo::fill(): APPLGRID requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   if ( id<0 ) { // fastNLO requested

#ifdef HAVE_FASTNLO

      // Fill fastNLO table

      //      std::cout << "### nnlo::fill(): Fill fastNLO table ..." << std::endl;

      // for (int i=0; i<nSubProcToBeFilled; i++) {
      ftableBase->fEvent.SetProcessId(0);
      ftableBase->fEvent.SetWeight(3.141);
      ftableBase->fEvent.SetX1(0.1);
      ftableBase->fEvent.SetX2(0.01);
      ftableBase->fScenario.SetObservable0(11.111);
      ftableBase->fScenario.SetObsScale1(10.);
      ftableBase->Fill(0);
      ftableBase->fEvent.Reset();
      // ftableBase->fEvent.SetProcessId(i);
      // ftableBase->fEvent.SetWeight(weights[i]/x1/x2);
      // ftableBase->fEvent.SetX1(x1);
      // ftableBase->fEvent.SetX2(x2);
      // ftableBase->fScenario.SetObservable0(coord);
      // ftableBase->fScenario.SetObsScale1(sqrt(pdfQ2));
      // ftableBase->Fill(0);
      // ftableBase->fEvent.Reset();
      // }

#else

      std::cerr << "### nnlo::fill(): fastNLO requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   //   std::cout << "### nnlo::fill(): End filling ..." << std::endl;

}



#ifdef HAVE_APPL
void nnlo::fill_appl_internal( appl::grid* _grid,
                                  const int& order,
                                  const double& x1, const double& x2, const double& Q2,
                                  const double& obs, const double* wgt ) {

  //  static int Nsubproc = _grid->subProcesses( inppar_.nloop );

  const lumi_pdf* pdf = dynamic_cast<const lumi_pdf*>(((const appl::grid*)_grid)->genpdf( inppar_.nloop ) );

  static std::vector<double> weight(pdf->size());

  static double  factor[4] = { 1, 1, 1, 1 };
  static bool   bfactor[4] = { false, false, false, false };

  //  int as_order = inppar_.njets + inppar_.nloop;
  int as_order = inppar_.njets + order;

  if ( !bfactor[order] ) {
    bfactor[order] = true;
    for ( int iorder=0 ; iorder<as_order ; iorder++ ) factor[order] *= 2*M_PI;
  }

  //  weight[0] = histonorm_.wthist * wt[0] * xjacob_.xjac1 / xi1 / xi2;
  for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] = wgt[i] / x1 / x2;

  for ( unsigned i=0 ; i<pdf->size() ; i++ ) weight[i] *= factor[order];

  if ( first_run ) _grid->fill_phasespace(  x1, x2, Q2, obs,  &weight[0], order );
  else             _grid->fill_grid(        x1, x2, Q2, obs,  &weight[0], order );

}
#endif





//
// fill the applgrid reference histogram.
//
void nnlo::fill_ref( const int& id, const double& obs, const double* wt ) {

#ifdef HAVE_APPL
  appl::grid* _grid = find( id );
  if ( _grid ) _grid->getReference()->Fill( obs, wt[0] );
#endif

#ifdef HAVE_FASTNLO
   //  std::cout << "To be done ..." << std::endl;
#endif

}





void Normalise(TH1D* h)
{
  for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) {
    double width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
    h->SetBinContent( ibin, h->GetBinContent(ibin)/width );
  }
  return;
}


//
// Write out the grids
//
void nnlo::term( const int& id ) {

   std::cout << "nnlo::term() called" << std::endl;

   if ( id>=0 ) { // APPLGRID requested

#ifdef HAVE_APPL

      // Write APPLGRID

      std::cout << "### nnlo::term(): Write APPLGRID ..." << std::endl;

      appl::grid* _grid = find( id );

      if ( _grid==0 ) return;

      std::string gridname = findname( id );

      if ( gridname=="" ) return;

      std::cout << "  iterations: " << runinfo_.itmax2 << "\n"
                << "  shots:      " << runinfo_.nshota << "\n"
                << "  total:      " << runinfo_.itmax2*runinfo_.nshota
                << std::endl;


      /// we need to scale by the bin width ???

      double nweights = runinfo_.itmax2*runinfo_.nshota;

      //  double nweights = icount;

      ///  _grids.back()->getReference()->Scale(1/(conv*binwidth[nh]/nweights));

      _grid->run() = nweights;

      /// hmmm, we divide by the bin width after the convolution
      /// if the weights are *already* divided by the bin width, we will
      /// need to scale *up* the weights in each bin by the bin width
      /// as a hack ( for fixed binning ) do that here ...

      TH1D* h = _grid->getReference();

      /// Fixme: fixed width bins bin width scaling - sort this out
      ///        properly later
      double width = 0;
      for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) {
         width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
	 h->SetBinContent( ibin, h->GetBinContent(ibin)/width );  
     }

      /// scale up the grid by the number of weights - vegas already
      /// includes a 1/nshots term for the integration
      /// the grid will always be stored "unweighted"
      (*_grid) *= nweights;

      /// scale up the reference histogram, since it already is correctly
      /// weighted after the filling, but we scale it down by the number
      /// of weights when we write it out, so that the histogram in the
      /// root file is the actual cross section
      ///  _grids.back()->getReference()->Scale( nweights );

      Normalise( _grid->getReference() );

      _grid->Write( gridname );

#else

      std::cerr << "### nnlo::term(): APPLGRID requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   if ( id<0 ) { // fastNLO requested

#ifdef HAVE_FASTNLO

      // Write fastNLO table

      std::cout << "### nnlo::term(): Write fastNLO table ..." << std::endl;
      unsigned long long int nEvents = 100;
      ftableBase->SetNumberOfEvents(nEvents);
      if (ftableBase->GetIsWarmup()) {
         std::cout << "### nnlo::term(): fastNLO warmup run finished ..." << std::endl;
         //         ftableBase->WriteTable();
      } else {
         std::cout << "### nnlo::term(): fastNLO run finished ..." << std::endl;
         //         ftableBase->SetFilename("NNLO-Test.tab");
      }
#else

      std::cerr << "### nnlo::term(): fastNLO requested, but not linked!" << std::endl;
      exit(1);

#endif

   }

   std::cout << "### nnlo::term(): End Term ..." << std::endl;

}
